@@ -90,31 +90,12 @@ gint at_util_call_compare(gconstpointer a, gconstpointer b)
 	return 0;
 }
 
-unsigned int at_util_alloc_next_id(unsigned int *id_list)
-{
-	unsigned int i;
-
-	for (i = 1; i < sizeof(unsigned int) * 8; i++) {
-		if (*id_list & (1 << i))
-			continue;
-
-		*id_list |= (1 << i);
-		return i;
-	}
-
-	return 0;
-}
-
-void at_util_release_id(unsigned int *id_list, unsigned int id)
-{
-	*id_list &= ~(1 << id);
-}
-
 GSList *at_util_parse_clcc(GAtResult *result)
 {
 	GAtResultIter iter;
 	GSList *l = NULL;
 	int id, dir, status, type;
+	ofono_bool_t mpty;
 	struct ofono_call *call;
 
 	g_at_result_iter_init(&iter, result);
@@ -135,7 +116,7 @@ GSList *at_util_parse_clcc(GAtResult *result)
 		if (!g_at_result_iter_next_number(&iter, &type))
 			continue;
 
-		if (!g_at_result_iter_skip_next(&iter))
+		if (!g_at_result_iter_next_number(&iter, &mpty))
 			continue;
 
 		if (g_at_result_iter_next_string(&iter, &str))
@@ -150,6 +131,7 @@ GSList *at_util_parse_clcc(GAtResult *result)
 		call->direction = dir;
 		call->status = status;
 		call->type = type;
+		call->mpty = mpty;
 		strncpy(call->phone_number.number, str,
 				OFONO_MAX_PHONE_NUMBER_LENGTH);
 		call->phone_number.type = number_type;
@@ -165,3 +147,99 @@ GSList *at_util_parse_clcc(GAtResult *result)
 	return l;
 }
 
+gboolean at_util_parse_reg_unsolicited(GAtResult *result, const char *prefix,
+					int *status,
+					int *lac, int *ci, int *tech)
+{
+	GAtResultIter iter;
+	int s;
+	int l = -1, c = -1, t = -1;
+	const char *str;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (g_at_result_iter_next(&iter, prefix) == FALSE)
+		return FALSE;
+		
+	if (g_at_result_iter_next_number(&iter, &s) == FALSE)
+		return FALSE;
+
+	if (g_at_result_iter_next_string(&iter, &str) == TRUE)
+		l = strtol(str, NULL, 16);
+	else
+		goto out;
+
+	if (g_at_result_iter_next_string(&iter, &str) == TRUE)
+		c = strtol(str, NULL, 16);
+	else
+		goto out;
+
+	g_at_result_iter_next_number(&iter, &t);
+
+out:
+	if (status)
+		*status = s;
+
+	if (lac)
+		*lac = l;
+
+	if (ci)
+		*ci = c;
+
+	if (tech)
+		*tech = t;
+
+	return TRUE;
+}
+
+gboolean at_util_parse_reg(GAtResult *result, const char *prefix,
+				int *mode, int *status,
+				int *lac, int *ci, int *tech)
+{
+	GAtResultIter iter;
+	int m, s;
+	int l = -1, c = -1, t = -1;
+	const char *str;
+
+	g_at_result_iter_init(&iter, result);
+
+	while (g_at_result_iter_next(&iter, prefix)) {
+		g_at_result_iter_next_number(&iter, &m);
+
+		/* Sometimes we get an unsolicited CREG/CGREG here, skip it */
+		if (g_at_result_iter_next_number(&iter, &s) == FALSE)
+			continue;
+
+		if (g_at_result_iter_next_string(&iter, &str) == TRUE)
+			l = strtol(str, NULL, 16);
+		else
+			goto out;
+
+		if (g_at_result_iter_next_string(&iter, &str) == TRUE)
+			c = strtol(str, NULL, 16);
+		else
+			goto out;
+
+		g_at_result_iter_next_number(&iter, &t);
+
+out:
+		if (mode)
+			*mode = m;
+
+		if (status)
+			*status = s;
+
+		if (lac)
+			*lac = l;
+
+		if (ci)
+			*ci = c;
+
+		if (tech)
+			*tech = t;
+
+		return TRUE;
+	}
+	
+	return FALSE;
+}
