@@ -36,9 +36,16 @@
 
 #include "gatchat.h"
 #include "gatresult.h"
+#include "simutil.h"
+#include "vendor.h"
 
 #include "atmodem.h"
 
+struct sim_data {
+	GAtChat *chat;
+	unsigned int vendor;
+};
+
 static const char *crsm_prefix[] = { "+CRSM:", NULL };
 
 static void at_crsm_info_cb(gboolean ok, GAtResult *result, gpointer user_data)
@@ -50,7 +57,7 @@ static void at_crsm_info_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	const guint8 *response;
 	gint sw1, sw2, len;
 	int flen, rlen;
-	enum ofono_sim_file_structure str;
+	int str;
 	unsigned char access[3];
 
 	dump_response("at_crsm_info_cb", ok, result);
@@ -63,54 +70,53 @@ static void at_crsm_info_cb(gboolean ok, GAtResult *result, gpointer user_data)
 
 	g_at_result_iter_init(&iter, result);
 
-	if (!g_at_result_iter_next(&iter, "+CRSM:")) {
-		CALLBACK_WITH_FAILURE(cb, -1, -1, -1, NULL, cbd->data);
-		return;
-	}
+	if (!g_at_result_iter_next(&iter, "+CRSM:"))
+		goto error;
 
 	g_at_result_iter_next_number(&iter, &sw1);
 	g_at_result_iter_next_number(&iter, &sw2);
 
 	if (!g_at_result_iter_next_hexstring(&iter, &response, &len) ||
 			(sw1 != 0x90 && sw1 != 0x91 && sw1 != 0x92) ||
-			(sw1 == 0x90 && sw2 != 0x00) ||
-			len < 14 || response[6] != 0x04 ||
-			(response[13] == 0x01 && len < 15)) {
-		CALLBACK_WITH_FAILURE(cb, -1, -1, -1, NULL, cbd->data);
-		return;
-	}
+			(sw1 == 0x90 && sw2 != 0x00))
+		goto error;
 
 	ofono_debug("crsm_info_cb: %02x, %02x, %i", sw1, sw2, len);
 
-	flen = (response[2] << 8) | response[3];
-	str = response[13];
-
-	access[0] = response[8];
-	access[1] = response[9];
-	access[2] = response[10];
-
-	if (str == 0x01 || str == 0x03)
-		rlen = response[14];
+	if (response[0] == 0x62)
+		ok = sim_parse_3g_get_response(response, len, &flen, &rlen,
+						&str, access, NULL);
 	else
-		rlen = 0;
+		ok = sim_parse_2g_get_response(response, len, &flen, &rlen,
+						&str, access);
+
+	if (!ok)
+		goto error;
 
 	cb(&error, flen, str, rlen, access, cbd->data);
+	return;
+
+error:
+	CALLBACK_WITH_FAILURE(cb, -1, -1, -1, NULL, cbd->data);
 }
 
 static void at_sim_read_info(struct ofono_sim *sim, int fileid,
 					ofono_sim_file_info_cb_t cb,
 					void *data)
 {
-	GAtChat *chat = ofono_sim_get_data(sim);
+	struct sim_data *sd = ofono_sim_get_data(sim);
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char buf[64];
 
 	if (!cbd)
 		goto error;
 
-	snprintf(buf, sizeof(buf), "AT+CRSM=192,%i,0,0,15", fileid);
+	snprintf(buf, sizeof(buf), "AT+CRSM=192,%i", fileid);
 
-	if (g_at_chat_send(chat, buf, crsm_prefix,
+	if (sd->vendor == OFONO_VENDOR_QUALCOMM_MSM)
+		strcat(buf, ",0,0,255"); /* Maximum possible length */
+
+	if (g_at_chat_send(sd->chat, buf, crsm_prefix,
 				at_crsm_info_cb, cbd, g_free) > 0)
 		return;
 
@@ -165,7 +171,7 @@ static void at_sim_read_binary(struct ofono_sim *sim, int fileid,
 					int start, int length,
 					ofono_sim_read_cb_t cb, void *data)
 {
-	GAtChat *chat = ofono_sim_get_data(sim);
+	struct sim_data *sd = ofono_sim_get_data(sim);
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char buf[64];
 
@@ -175,7 +181,7 @@ static void at_sim_read_binary(struct ofono_sim *sim, int fileid,
 	snprintf(buf, sizeof(buf), "AT+CRSM=176,%i,%i,%i,%i", fileid,
 			start >> 8, start & 0xff, length);
 
-	if (g_at_chat_send(chat, buf, crsm_prefix,
+	if (g_at_chat_send(sd->chat, buf, crsm_prefix,
 				at_crsm_read_cb, cbd, g_free) > 0)
 		return;
 
@@ -190,7 +196,7 @@ static void at_sim_read_record(struct ofono_sim *sim, int fileid,
 					int record, int length,
 					ofono_sim_read_cb_t cb, void *data)
 {
-	GAtChat *chat = ofono_sim_get_data(sim);
+	struct sim_data *sd = ofono_sim_get_data(sim);
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char buf[64];
 
@@ -200,7 +206,7 @@ static void at_sim_read_record(struct ofono_sim *sim, int fileid,
 	snprintf(buf, sizeof(buf), "AT+CRSM=178,%i,%i,4,%i", fileid,
 			record, length);
 
-	if (g_at_chat_send(chat, buf, crsm_prefix,
+	if (g_at_chat_send(sd->chat, buf, crsm_prefix,
 				at_crsm_read_cb, cbd, g_free) > 0)
 		return;
 
@@ -254,7 +260,7 @@ static void at_sim_update_binary(struct ofono_sim *sim, int fileid,
 					const unsigned char *value,
 					ofono_sim_write_cb_t cb, void *data)
 {
-	GAtChat *chat = ofono_sim_get_data(sim);
+	struct sim_data *sd = ofono_sim_get_data(sim);
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char *buf = g_try_new(char, 36 + length * 2);
 	int len, ret;
@@ -266,9 +272,9 @@ static void at_sim_update_binary(struct ofono_sim *sim, int fileid,
 			start >> 8, start & 0xff, length);
 
 	for (; length; length--)
-		len += sprintf(buf + len, "%02hhx", *value++);
+		len += sprintf(buf + len, "%02hhX", *value++);
 
-	ret = g_at_chat_send(chat, buf, crsm_prefix,
+	ret = g_at_chat_send(sd->chat, buf, crsm_prefix,
 				at_crsm_update_cb, cbd, g_free);
 
 	g_free(buf);
@@ -288,7 +294,7 @@ static void at_sim_update_record(struct ofono_sim *sim, int fileid,
 					const unsigned char *value,
 					ofono_sim_write_cb_t cb, void *data)
 {
-	GAtChat *chat = ofono_sim_get_data(sim);
+	struct sim_data *sd = ofono_sim_get_data(sim);
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char *buf = g_try_new(char, 36 + length * 2);
 	int len, ret;
@@ -300,9 +306,9 @@ static void at_sim_update_record(struct ofono_sim *sim, int fileid,
 			record, length);
 
 	for (; length; length--)
-		len += sprintf(buf + len, "%02hhx", *value++);
+		len += sprintf(buf + len, "%02hhX", *value++);
 
-	ret = g_at_chat_send(chat, buf, crsm_prefix,
+	ret = g_at_chat_send(sd->chat, buf, crsm_prefix,
 				at_crsm_update_cb, cbd, g_free);
 
 	g_free(buf);
@@ -321,7 +327,7 @@ static void at_sim_update_cyclic(struct ofono_sim *sim, int fileid,
 					int length, const unsigned char *value,
 					ofono_sim_write_cb_t cb, void *data)
 {
-	GAtChat *chat = ofono_sim_get_data(sim);
+	struct sim_data *sd = ofono_sim_get_data(sim);
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char *buf = g_try_new(char, 36 + length * 2);
 	int len, ret;
@@ -332,9 +338,9 @@ static void at_sim_update_cyclic(struct ofono_sim *sim, int fileid,
 	len = sprintf(buf, "AT+CRSM=220,%i,0,3,%i,", fileid, length);
 
 	for (; length; length--)
-		len += sprintf(buf + len, "%02hhx", *value++);
+		len += sprintf(buf + len, "%02hhX", *value++);
 
-	ret = g_at_chat_send(chat, buf, crsm_prefix,
+	ret = g_at_chat_send(sd->chat, buf, crsm_prefix,
 				at_crsm_update_cb, cbd, g_free);
 
 	g_free(buf);
@@ -381,13 +387,13 @@ static void at_cimi_cb(gboolean ok, GAtResult *result, gpointer user_data)
 static void at_read_imsi(struct ofono_sim *sim, ofono_sim_imsi_cb_t cb,
 			void *data)
 {
-	GAtChat *chat = ofono_sim_get_data(sim);
+	struct sim_data *sd = ofono_sim_get_data(sim);
 	struct cb_data *cbd = cb_data_new(cb, data);
 
 	if (!cbd)
 		goto error;
 
-	if (g_at_chat_send(chat, "AT+CIMI", NULL,
+	if (g_at_chat_send(sd->chat, "AT+CIMI", NULL,
 				at_cimi_cb, cbd, g_free) > 0)
 		return;
 
@@ -469,13 +475,13 @@ static void at_cpin_cb(gboolean ok, GAtResult *result, gpointer user_data)
 static void at_pin_query(struct ofono_sim *sim, ofono_sim_passwd_cb_t cb,
 			void *data)
 {
-	GAtChat *chat = ofono_sim_get_data(sim);
+	struct sim_data *sd = ofono_sim_get_data(sim);
 	struct cb_data *cbd = cb_data_new(cb, data);
 
 	if (!cbd)
 		goto error;
 
-	if (g_at_chat_send(chat, "AT+CPIN?", NULL,
+	if (g_at_chat_send(sd->chat, "AT+CPIN?", NULL,
 				at_cpin_cb, cbd, g_free) > 0)
 		return;
 
@@ -502,7 +508,7 @@ static void at_lock_unlock_cb(gboolean ok, GAtResult *result,
 static void at_pin_send(struct ofono_sim *sim, const char *passwd,
 			ofono_sim_lock_unlock_cb_t cb, void *data)
 {
-	GAtChat *chat = ofono_sim_get_data(sim);
+	struct sim_data *sd = ofono_sim_get_data(sim);
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char buf[64];
 	int ret;
@@ -512,7 +518,8 @@ static void at_pin_send(struct ofono_sim *sim, const char *passwd,
 
 	snprintf(buf, sizeof(buf), "AT+CPIN=\"%s\"", passwd);
 
-	ret = g_at_chat_send(chat, buf, NULL, at_lock_unlock_cb, cbd, g_free);
+	ret = g_at_chat_send(sd->chat, buf, NULL,
+				at_lock_unlock_cb, cbd, g_free);
 
 	memset(buf, 0, sizeof(buf));
 
@@ -530,7 +537,7 @@ static void at_pin_send_puk(struct ofono_sim *sim, const char *puk,
 				const char *passwd,
 				ofono_sim_lock_unlock_cb_t cb, void *data)
 {
-	GAtChat *chat = ofono_sim_get_data(sim);
+	struct sim_data *sd = ofono_sim_get_data(sim);
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char buf[64];
 	int ret;
@@ -540,7 +547,8 @@ static void at_pin_send_puk(struct ofono_sim *sim, const char *puk,
 
 	snprintf(buf, sizeof(buf), "AT+CPIN=\"%s\",\"%s\"", puk, passwd);
 
-	ret = g_at_chat_send(chat, buf, NULL, at_lock_unlock_cb, cbd, g_free);
+	ret = g_at_chat_send(sd->chat, buf, NULL,
+				at_lock_unlock_cb, cbd, g_free);
 
 	memset(buf, 0, sizeof(buf));
 
@@ -570,7 +578,7 @@ static void at_pin_enable(struct ofono_sim *sim,
 				int enable, const char *passwd,
 				ofono_sim_lock_unlock_cb_t cb, void *data)
 {
-	GAtChat *chat = ofono_sim_get_data(sim);
+	struct sim_data *sd = ofono_sim_get_data(sim);
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char buf[64];
 	int ret;
@@ -585,7 +593,8 @@ static void at_pin_enable(struct ofono_sim *sim,
 	snprintf(buf, sizeof(buf), "AT+CLCK=\"%s\",%i,\"%s\"",
 			at_clck_cpwd_fac[passwd_type], enable ? 1 : 0, passwd);
 
-	ret = g_at_chat_send(chat, buf, NULL, at_lock_unlock_cb, cbd, g_free);
+	ret = g_at_chat_send(sd->chat, buf, NULL,
+				at_lock_unlock_cb, cbd, g_free);
 
 	memset(buf, 0, sizeof(buf));
 
@@ -604,7 +613,7 @@ static void at_change_passwd(struct ofono_sim *sim,
 			const char *old, const char *new,
 			ofono_sim_lock_unlock_cb_t cb, void *data)
 {
-	GAtChat *chat = ofono_sim_get_data(sim);
+	struct sim_data *sd = ofono_sim_get_data(sim);
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char buf[64];
 	int ret;
@@ -620,7 +629,8 @@ static void at_change_passwd(struct ofono_sim *sim,
 	snprintf(buf, sizeof(buf), "AT+CPWD=\"%s\",\"%s\",\"%s\"",
 			at_clck_cpwd_fac[passwd_type], old, new);
 
-	ret = g_at_chat_send(chat, buf, NULL, at_lock_unlock_cb, cbd, g_free);
+	ret = g_at_chat_send(sd->chat, buf, NULL,
+				at_lock_unlock_cb, cbd, g_free);
 
 	memset(buf, 0, sizeof(buf));
 
@@ -669,7 +679,7 @@ static void at_pin_query_enabled(struct ofono_sim *sim,
 				enum ofono_sim_password_type passwd_type,
 				ofono_sim_locked_cb_t cb, void *data)
 {
-	GAtChat *chat = ofono_sim_get_data(sim);
+	struct sim_data *sd = ofono_sim_get_data(sim);
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char buf[64];
 	unsigned int len = sizeof(at_clck_cpwd_fac) / sizeof(*at_clck_cpwd_fac);
@@ -683,7 +693,7 @@ static void at_pin_query_enabled(struct ofono_sim *sim,
 	snprintf(buf, sizeof(buf), "AT+CLCK=\"%s\",2",
 			at_clck_cpwd_fac[passwd_type]);
 
-	if (g_at_chat_send(chat, buf, NULL,
+	if (g_at_chat_send(sd->chat, buf, NULL,
 				at_lock_status_cb, cbd, g_free) > 0)
 		return;
 
@@ -694,6 +704,83 @@ error:
 	CALLBACK_WITH_FAILURE(cb, -1, data);
 }
 
+static void at_csim_envelope_cb(gboolean ok, GAtResult *result,
+		gpointer user_data)
+{
+	struct cb_data *cbd = user_data;
+	GAtResultIter iter;
+	ofono_sim_read_cb_t cb = cbd->cb;
+	struct ofono_error error;
+	const guint8 *response;
+	gint rlen, len;
+
+	dump_response("at_csim_envelope_cb", ok, result);
+	decode_at_error(&error, g_at_result_final_response(result));
+
+	if (!ok)
+		goto error;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "+CSIM:"))
+		goto error;
+
+	if (!g_at_result_iter_next_number(&iter, &rlen))
+		goto error;
+
+	if (!g_at_result_iter_next_hexstring(&iter, &response, &len))
+		goto error;
+
+	if (rlen != len * 2 || len < 2 ||
+			response[len - 2] != 0x90 || response[len - 1] != 0)
+		goto error;
+
+	ofono_debug("csim_envelope_cb: %i", len);
+
+	cb(&error, response, len - 2, cbd->data);
+	return;
+
+error:
+	CALLBACK_WITH_FAILURE(cb, NULL, 0, cbd->data);
+}
+
+static void at_sim_envelope(struct ofono_sim *sim, int length,
+				const guint8 *command,
+				ofono_sim_read_cb_t cb, void *data)
+{
+	struct sim_data *sd = ofono_sim_get_data(sim);
+	struct cb_data *cbd = cb_data_new(cb, data);
+	char *buf = g_try_new(char, 64 + length * 2);
+	int len, ret;
+
+	if (!cbd || !buf)
+		goto error;
+
+	len = sprintf(buf, "AT+CSIM=%i,A0C20000%02hhX",
+			10 + length * 2, length);
+
+	for (; length; length--)
+		len += sprintf(buf + len, "%02hhX", *command++);
+
+	ret = g_at_chat_send(sd->chat, buf, crsm_prefix,
+				at_csim_envelope_cb, cbd, g_free);
+
+	g_free(buf);
+	buf = NULL;
+
+	if (ret > 0)
+		return;
+
+error:
+	if (buf)
+		g_free(buf);
+
+	if (cbd)
+		g_free(cbd);
+
+	CALLBACK_WITH_FAILURE(cb, NULL, 0, data);
+}
+
 static gboolean at_sim_register(gpointer user)
 {
 	struct ofono_sim *sim = user;
@@ -707,8 +794,13 @@ static int at_sim_probe(struct ofono_sim *sim, unsigned int vendor,
 				void *data)
 {
 	GAtChat *chat = data;
+	struct sim_data *sd;
 
-	ofono_sim_set_data(sim, chat);
+	sd = g_new0(struct sim_data, 1);
+	sd->chat = chat;
+	sd->vendor = vendor;
+
+	ofono_sim_set_data(sim, sd);
 	g_idle_add(at_sim_register, sim);
 
 	return 0;
@@ -716,6 +808,11 @@ static int at_sim_probe(struct ofono_sim *sim, unsigned int vendor,
 
 static void at_sim_remove(struct ofono_sim *sim)
 {
+	struct sim_data *sd = ofono_sim_get_data(sim);
+
+	ofono_sim_set_data(sim, NULL);
+
+	g_free(sd);
 }
 
 static struct ofono_sim_driver driver = {
@@ -736,6 +833,7 @@ static struct ofono_sim_driver driver = {
 	.lock			= at_pin_enable,
 	.change_passwd		= at_change_passwd,
 	.query_locked		= at_pin_query_enabled,
+	.envelope		= at_sim_envelope,
 };
 
 void at_sim_init()
