@@ -41,7 +41,7 @@
 #define uninitialized_var(x) x = x
 
 #define SMS_BACKUP_MODE 0600
-#define SMS_BACKUP_PATH STORAGEDIR "/%s/sms"
+#define SMS_BACKUP_PATH STORAGEDIR "/%s/sms_assembly"
 #define SMS_BACKUP_PATH_DIR SMS_BACKUP_PATH "/%s-%i-%i"
 #define SMS_BACKUP_PATH_FILE SMS_BACKUP_PATH_DIR "/%03i"
 
@@ -805,7 +805,7 @@ static gboolean decode_submit_report(const unsigned char *pdu, int len,
 	int offset = 0;
 	unsigned char octet;
 	gboolean udhi;
-	guint8 fcs;
+	guint8 uninitialized_var(fcs);
 	guint8 pi;
 	struct sms_scts *scts;
 	guint8 pid = 0;
@@ -1003,8 +1003,9 @@ static gboolean decode_status_report(const unsigned char *pdu, int len,
 	if (out->status_report.pi & 0x02) {
 		if (!next_octet(pdu, len, &offset, &out->status_report.dcs))
 			return FALSE;
-	} else
+	} else {
 		out->status_report.dcs = 0;
+	}
 
 	if (out->status_report.pi & 0x04) {
 		int expected;
@@ -1096,7 +1097,7 @@ static gboolean decode_deliver_report(const unsigned char *pdu, int len,
 	int offset = 0;
 	unsigned char octet;
 	gboolean udhi;
-	guint8 fcs;
+	guint8 uninitialized_var(fcs);
 	guint8 pi;
 	guint8 pid = 0;
 	guint8 dcs = 0;
@@ -1125,8 +1126,9 @@ static gboolean decode_deliver_report(const unsigned char *pdu, int len,
 
 		if (!next_octet(pdu, len, &offset, &octet))
 			return FALSE;
-	} else
+	} else {
 		out->type = SMS_TYPE_DELIVER_REPORT_ACK;
+	}
 
 	pi = octet & 0x07;
 
@@ -1776,8 +1778,9 @@ const char *sms_address_to_string(const struct sms_address *addr)
 				addr->address[0] != '+') {
 		buffer[0] = '+';
 		strcpy(buffer + 1, addr->address);
-	} else
+	} else {
 		strcpy(buffer, addr->address);
+	}
 
 	return buffer;
 }
@@ -2133,6 +2136,9 @@ char *sms_decode_text(GSList *sms_list)
 
 		udl_in_bytes = sms_udl_in_bytes(udl, dcs);
 
+		if (udl_in_bytes == taken)
+			continue;
+
 		if (charset == SMS_CHARSET_7BIT) {
 			unsigned char buf[160];
 			long written;
@@ -2140,9 +2146,11 @@ char *sms_decode_text(GSList *sms_list)
 			guint8 single_shift = 0;
 			int max_chars = sms_text_capacity_gsm(udl, taken);
 
-			unpack_7bit_own_buf(ud + taken, udl_in_bytes - taken,
+			if (unpack_7bit_own_buf(ud + taken,
+						udl_in_bytes - taken,
 						taken, FALSE, max_chars,
-						&written, 0, buf);
+						&written, 0, buf) == NULL)
+				continue;
 
 			/* Take care of improperly split fragments */
 			if (buf[written-1] == 0x1b)
@@ -2185,11 +2193,6 @@ char *sms_decode_text(GSList *sms_list)
 
 	utf8 = g_string_free(str, FALSE);
 
-	if (strlen(utf8) == 0) {
-		g_free(utf8);
-		return NULL;
-	}
-
 	return utf8;
 }
 
@@ -3092,8 +3095,9 @@ char *cbs_decode_text(GSList *cbs_list, char *iso639_lang)
 						NULL, 0,
 						(unsigned char *)iso639_lang);
 			iso639_lang[2] = '\0';
-		} else
+		} else {
 			iso639_2_from_language(lang, iso639_lang);
+		}
 	}
 
 	buf = g_new(unsigned char, bufsize);
@@ -3509,7 +3513,7 @@ out:
 	return TRUE;
 }
 
-static GSList *cbs_optimize_ranges(GSList *ranges)
+GSList *cbs_optimize_ranges(GSList *ranges)
 {
 	struct cbs_topic_range *range;
 	unsigned char bitmap[125];
@@ -3666,3 +3670,23 @@ char *cbs_topic_ranges_to_string(GSList *ranges)
 
 	return ret;
 }
+
+static gint cbs_topic_compare(gconstpointer a, gconstpointer b)
+{
+	const struct cbs_topic_range *range = a;
+	unsigned short topic = GPOINTER_TO_UINT(b);
+
+	if (topic >= range->min && topic <= range->max)
+		return 0;
+
+	return 1;
+}
+
+gboolean cbs_topic_in_range(unsigned int topic, GSList *ranges)
+{
+	if (!ranges)
+		return FALSE;
+
+	return g_slist_find_custom(ranges, GUINT_TO_POINTER(topic),
+					cbs_topic_compare) != NULL;
+}
